#!/usr/bin/perl

#
# dbcolsdecimate.pm
# Copyright (C) 2023 by John Heidemann <johnh@isi.edu>
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#

package Fsdb::Filter::dbcolsdecimate;

=head1 NAME

dbcolsdecimate - prune columns, keeping "important" points

=head1 SYNOPSIS

    dbcolsdecimate [-p PRECISION] column1 [column2...]

=head1 DESCRIPTION

For each of the given columns, prune it back to show changes
with at most PRECISION fraction of total range change (default: 0.01).
This tool is designed for reducing the actual data in a graph
while keeping it visually identical.

With multiple columns, major changes in any column are shown.

Our goal is to output an identical plot, with fewer points if we can.
This goal differs from and is easier than
prior published work that has the goal of
the number of points by a known factor, or to a constant number,
while preserving as much fidelity as possible.

This program stores a complete copy of the input data on disk
temporarily.

=head1 OPTIONS

=over 4

=item B<--precision> P or B<-p> P

Set the precision of how large a fraction of the total range
should be presereved.
Default: 0.01.

=item B<-T TmpDir>

where to put tmp files.
Also uses environment variable TMPDIR, if -T is 
not specified.
Default is /tmp.

=back

=for comment
begin_standard_fsdb_options

This module also supports the standard fsdb options:

=over 4

=item B<-d>

Enable debugging output.

=item B<-v>

Enable verbose output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Fsdb::IO or Fsdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Fsdb::IO or Fsdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Fsdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_fsdb_options


=head1 SAMPLE USAGE

=head2 Input:

xxx

=head2 Command:

    cat xxx | dbcolsdecimate xxx

=head2 Output:

xxx


=head1 SEE ALSO

L<Fsdb>,
L<dbcolmovingstats>.


=head1 CLASS FUNCTIONS

=cut

@ISA = qw(Fsdb::Filter);
$VERSION = 2.0;

use strict;
use Pod::Usage;
use Carp;

use Fsdb::Filter;
use Fsdb::IO::Reader;
use Fsdb::IO::Writer;
use Fsdb::Support qw($is_numeric_regexp);
use Fsdb::Support::NamedTmpfile;
use Fsdb::Filter::dbpipeline qw(dbpipeline_open2 dbpipeline_close2_hash dbcolstats);


=head2 new

    $filter = new Fsdb::Filter::dbcolsdecimate(@arguments);

Create a new dbcolsdecimate object, taking command-line arguments.

=cut

sub new ($@) {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my($self) = @_;
    $self->SUPER::set_defaults();
    $self->{_columns} = [];
    $self->{_include_non_numeric} = undef;
    $self->{_precision} = 0.01;
    $self->{_verbose} = 0;
    $self->set_default_tmpdir;
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse command-line arguments.

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my(@argv) = @_;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'a|include-non-numeric!' => \$self->{_include_non_numeric},
	'autorun!' => \$self->{_autorun},
	'd|debug+' => \$self->{_debug},
	'v|verbose+' => \$self->{_verbose},
	'i|input=s' => sub { $self->parse_io_option('input', @_); },
	'log!' => \$self->{_logprog},
	'p|precision=f' => \$self->{_precision},
	'T|tmpdir|tempdir=s' => \$self->{_tmpdir},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	) or pod2usage(2);
    push (@{$self->{_columns}}, @argv);
}

=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup($) {
    my($self) = @_;

    croak($self->{_prog} . ": include-non-numeric is not currently supported.\n")
	if ($self->{_include_non_numeric});
    $self->finish_io_option('input', -comment_handler => $self->create_pass_comments_sub);
    croak($self->{_prog} . ": at least one column must be specified to decimate.\n")
	if ($#{$self->{_columns}} < 0);
    my %columns_processed;
    foreach my $i (0..$#{$self->{_columns}}) {
	my $column = $self->{_columns}[$i];
	croak($self->{_prog} . ": column $column is double-listed as an input column (not allowed).\n")
	    if (defined($columns_processed{$column}));
	$columns_processed{$column} = 1;
	$self->{_colis}[$i] = $self->{_in}->col_to_i($column);
	croak($self->{_prog} . ": column $column does not exist in the input stream.\n")
	    if (!defined($self->{_colis}[$i]));
    };
    $self->finish_io_option('output', -clone => $self->{_in});
};


=head2 run

    $filter->run();

Internal: run over each rows.

=cut
sub run ($) {
    my($self) = @_;

    #
    # First, read data and save it to a file,
    # recording min and max as we go.
    # 
    $self->{_copy_filename} = Fsdb::Support::NamedTmpfile::alloc($self->{_tmpdir});
    my $copy_writer = new Fsdb::IO::Writer(-file => $self->{_copy_filename},
			-clone => $self->{_in});

    my $read_fastpath_sub = $self->{_in}->fastpath_sub();
    my $copy_fastpath_sub = $copy_writer->fastpath_sub();

    # and take stats
    my(@mins, @maxes);
    my $colis_aref = $self->{_colis};
    my $ncolis = $#$colis_aref;
    my $fref;
    while ($fref = &$read_fastpath_sub()) {
	# copy and send to stats
	$copy_writer->write_rowobj($fref);
        foreach (0..$ncolis) {
            my($x) = $fref->[$colis_aref->[$_]];
            if ($x !~ /$is_numeric_regexp/) {
                if ($self->{_include_non_numeric}) {
                    $x = 0;
                } else {
                    next;
                };
            };
            if (!defined($mins[$_])) {
                $mins[$_] = $maxes[$_] = $x;
            } else {
                $mins[$_] = $x if ($x < $mins[$_]);
                $maxes[$_] = $x if ($x > $maxes[$_]);
            };
        };
    };
    # close up both
    $copy_writer->close;

    #
    # make a plan!
    #
    my(@ranges);
    my(@thresholds);
    foreach (0..$ncolis) {
        croak($self->{_prog} . ": column " . $self->{_columns}[$_] . " has no data, giving up.\n")
            if (!defined($mins[$_]));
        $ranges[$_] = $maxes[$_] - $mins[$_];
        $thresholds[$_] = $ranges[$_] * $self->{_precision};
    };

    #
    # Now read back the data
    # and decimate as we go.
    #
    $self->{_in}->close;
    $self->{_in} = new Fsdb::IO::Reader(-file => $self->{_copy_filename},
		    -comment_handler => $self->create_pass_comments_sub);
	    
    $read_fastpath_sub = $self->{_in}->fastpath_sub(); # regenerate with copy stream
    my $write_fastpath_sub = $self->{_out}->fastpath_sub();

    my($lfref_kept) = undef;
    my($lfref_immediate) = undef;
    my $rows_output = 0;
    my $rows_skipped = 0;
    while ($fref = &$read_fastpath_sub()) {
        my $interesting = !defined($lfref_kept);  # always do the first
        if (!$interesting) {
            foreach (0..$ncolis) {
                my $i = $colis_aref->[$_];
                if (abs($fref->[$i] - $lfref_kept->[$i]) > $thresholds[$_]) {
                    # big move, so put something out
                    $interesting = 1;
                    last;
                };
            };
        };
        if ($interesting) {
            #
            # Time to output.
            #
            # First try to do a step, if necessary.
            #
            if ($lfref_immediate) {
                &$write_fastpath_sub($lfref_immediate);
                $rows_output++;
                $rows_skipped--;   # unskip it!
                $lfref_immediate = undef;
            };
            # now us
            &$write_fastpath_sub($fref);
            $lfref_kept = $fref;
            $rows_output++;
            die "assertion failed" if (defined($lfref_immediate));
        } else {
            #
            # Time to skip.
            #
            $rows_skipped++;
            $lfref_immediate = $fref;
        };
    };
    #
    # Always output the last one, even if uninteresting.
    #
    die "assertion failed" if (defined($fref));
    if ($lfref_immediate) {
        &$write_fastpath_sub($lfref_immediate);
        $rows_output++;
        $rows_skipped--;   # unskip it!
        $lfref_immediate = undef;
    };
    my $rows_total = $rows_output + $rows_skipped;
    my $fraction_output = sprintf("%0.4f", ($rows_output * 1.0 / $rows_total));
    $self->{_out}->write_comment(" output $rows_output of $rows_total ($fraction_output)") if ($self->{_verbose});
};

=head1 AUTHOR and COPYRIGHT

Copyright (C) 2023 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;
